/*
 * Decompresses some PeX 0.99 files as well as all known public versions of Continuum.
 * Written by Snrrrub (snrrrub@yahoo.com): April 16, 2004
 * Original DeX code (c) by Chafe 2000
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commdlg.h>
#include <stdio.h>
#include <stdlib.h>
#include "Decompress.h"

// Returns the length of the file
long fileLength(const char * filename)
{
	FILE * fp = fopen(filename, "rb");
	if(!fp)
		return 0;
	fseek(fp, 0, SEEK_END);
	long len = ftell(fp);
	fclose(fp);
	return len;
}

// Reads a file into a new buffer. The buffer needs to be freed when no
// longer needed. Returns 0 if file cannot be opened.
void * readFile(const char * filename)
{
	long length = fileLength(filename);
	if(length == 0)
		return 0;

	FILE * fp = fopen(filename, "rb");
	if(!fp)
		return 0;
	void * data = malloc(length);
	fread(data, 1, length, fp);
	fclose(fp);
	return data;
}

// Writes the data to a file. Returns true if successful, false if not.
bool writeFile(const char * filename, const void * data, long length)
{
	bool ret = false;
	FILE * fp = fopen(filename, "wb");
	if(fp)
	{
		ret = (fwrite(data, 1, length, fp) == length);
		fclose(fp);
	}
	return ret;
}

struct Fixup
{
	long index;							// the offset in the binary to fix up
	unsigned char value;				// the new value at that offset
};
struct FixupTable
{
	long nFixups;						// # of fixup items in the pointer
	Fixup * fixup;						// A bunch of fixups for this version
};

Fixup ctm37Fixup = { 0x1FBE6, 0xEB };
Fixup ctm38Fixup[] = {
	{ 0x208F6, 0xEB },					// Disable "virus" check

	{ 0x32921, 0x33 },					// XOR checksum (corrupted executable)
	{ 0x32922, 0xC9 },
	{ 0x32923, 0x90 },

	{ 0x46B74, 0x8B },					// Big checksum (exit)
	{ 0x46B75, 0xC7 },
	{ 0x46B76, 0x3B },
	{ 0x46B77, 0xC7 },
	{ 0x46B78, 0x90 },
	{ 0x46B79, 0x90 },
	{ 0x46B7A, 0x90 },

	{ 0x015F5, 0x66 },					// Checksum result to ffd7778 used as a pointer
	{ 0x015F6, 0x89 },
	{ 0x015F7, 0x96 },
	{ 0x015F8, 0xB0 },
	{ 0x015F9, 0x04 },
	{ 0x015FA, 0x00 },
	{ 0x015FB, 0x00 },

	{ 0x3F55B, 0x66 },					// Checksum result to fffb3480 used as a pointer
	{ 0x3F55C, 0x89 },
	{ 0x3F55D, 0x4E },
	{ 0x3F55E, 0x6C },
	{ 0x3F55F, 0x90 },
	{ 0x3F560, 0x90 },
	{ 0x3F561, 0x90 },

	{ 0x268D0, 0xC7 },
	{ 0x268D1, 0x86 },
	{ 0x268D2, 0xD8 },
	{ 0x268D3, 0x27 },
	{ 0x268D4, 0x00 },
	{ 0x268D5, 0x00 },
	{ 0x268D6, 0x00 },
	{ 0x268D7, 0x00 },
	{ 0x268D8, 0x00 },
	{ 0x268D9, 0x00 },
	{ 0x268DA, 0x90 },

	{ 0x2E906, 0x8D },
	{ 0x2E907, 0x42 },
	{ 0x2E908, 0x1C },
	{ 0x2E909, 0x90 },
	{ 0x2E90A, 0x90 },
	{ 0x2E90B, 0x90 },
	{ 0x2E90C, 0x90 },

	{ 0x484A7, 0x66 },
	{ 0x484A8, 0x89 },
	{ 0x484A9, 0x4A },
	{ 0x484AA, 0x60 },
	{ 0x484AB, 0x90 },
	{ 0x484AC, 0x90 },
	{ 0x484AD, 0x90 }
};

FixupTable binaryFixups[] = {
	{  0, 0 },							// Standard PeX .99
	{  0, 0 },							// Ctm 0.35 and Ctm 0.36
	{  1, &ctm37Fixup },				// Ctm 0.37
	{ 50, ctm38Fixup },					// Ctm 0.38
	{  0, 0 }							// Ctm 0.39 pr 1 (patching disabled since I don't have the offset yet)
};

extern "C" int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	char fileName[1024];

	OPENFILENAME ofn;
	memset(&ofn, 0, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(OPENFILENAME);
	ofn.hInstance = GetModuleHandle(0);
	ofn.nFilterIndex = 1;
	ofn.lpstrFilter = "Executables (*.exe)\0*.exe\0\0";
	ofn.lpstrFile = fileName;
	ofn.nMaxFile = 1024;
	ofn.Flags = OFN_FILEMUSTEXIST | OFN_OVERWRITEPROMPT| OFN_HIDEREADONLY | OFN_LONGNAMES | OFN_PATHMUSTEXIST;
	ofn.lpstrDefExt = ".exe";

	// Try to open a file specified by the user
	if(GetOpenFileName(&ofn) == 0)
	{
		MessageBox(0, "Unpacking cancelled.", "Cancelled", MB_OK);
		return 1;
	}

	// read the file, make sure it's a PE
	long length = fileLength(fileName);
	m_inputFile = readFile(fileName);
	PEOffset = getPEOffset(m_inputFile, length);
	if(PEOffset)
	{
		// Find all the addresses and stuff
		SectTable = getSectionTable(PEOffset);
		PEXSection = (char *)m_inputFile + getPhysicalOffset(findSectionEntryWithRVA(SectTable, getEntryPointRVA(PEOffset)));
		Version = findVersion(generateChecksum(PEXSection));

		// We've recognized this version!
		if(Version != 0)
		{
			// Unpack it
			unpack();

			// Patch it
			FixupTable * thisFixup = &binaryFixups[Version->index];
			for(int i = 0; i < thisFixup->nFixups; ++i)
				((unsigned char *)m_outputFile)[thisFixup->fixup[i].index] = thisFixup->fixup[i].value;

			// Tell the user
			char buffer[1024];
			sprintf(buffer, "Successfully unpacked executable by recognizing %s.", Version->versionName);
			MessageBox(0, buffer, "Success!", MB_OK | MB_ICONINFORMATION);

			// Save it
			fileName[0] = 0;
			if(GetSaveFileName(&ofn) != 0)
				writeFile(fileName, m_outputFile, m_outputFileSize);
			else
				MessageBox(0, "Operation cancelled. File not saved.", "Cancelled", MB_OK | MB_ICONERROR);
		}
		else
			MessageBox(0, "Sorry, the specified executable is not packed using\na known algorithm.", "Error", MB_OK | MB_ICONERROR);
	}
	else
		MessageBox(0, "The file you have selected is not\nan executable.", "Error", MB_OK | MB_ICONERROR);

	FreeAllStuff();
	free(m_inputFile);

	return 0;
}
